
MODULE wadlmt
   !!==============================================================================
   !!                       ***  MODULE  wadlmt  ***
   !! compute the flux limiter for water depth positivity in !wetting/drying case
   !!==============================================================================
   !! History :   
   !!  NEMO      3.5  ! 2014-01  ((H.Liu)  Original code
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   wad_lmt    : Compute the horizontal flux limiter and the limited velocity
   !!                when wetting and drying happens 
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE sbc_oce, ONLY : ln_rnf   ! surface boundary condition: ocean
   USE sbcrnf          ! river runoff 
   USE cla             ! cross land advection             (cla_div routine)
   USE in_out_manager  ! I/O manager
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)
   USE lib_mpp         ! MPP library
   USE wrk_nemo        ! Memory Allocation
   USE timing          ! Timing

   IMPLICIT NONE
   PRIVATE


   PUBLIC   wad_lmt    ! routine called by step.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
CONTAINS

   SUBROUTINE wad_lmt( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE wad_lmt  ***
      !!                    
      !! ** Purpose :   generate flux limiters for wetting/drying
      !!
      !! ** Method  : - Prevent negative depth occurring (Not ready for Agrif) 
      !!
      !! ** Action  : - calculate flux limiter and W/D flag
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt       ! ocean time-step index
      !
      INTEGER  ::   ji, jj, jk, jk1     ! dummy loop indices
      INTEGER  ::   z2dt, zflag         ! local scalar
      REAL(wp) ::   zcoef, zdep1, zdep2 ! local scalars
      REAL(wp) ::   zzflxp, zzflxn      ! local scalars
      REAL(wp) ::   ztmp                ! local scalars
      INTEGER,  POINTER,  DIMENSION(:,:) ::   iwdflg                   !: W/D cell flag
      REAL(wp), POINTER,  DIMENSION(:,:) ::   zwdlmtu, zwdlmtv         !: W/D flux limiters
      REAL(wp), POINTER,  DIMENSION(:,:) ::   zflxp,  zflxn            ! specific 2D workspace
      REAL(wp), POINTER,  DIMENSION(:,:) ::   zflxu,  zflxv            ! specific 2D workspace
      REAL(wp), POINTER,  DIMENSION(:,:) ::   zflxu1, zflxv1           ! specific 2D workspace

      !!----------------------------------------------------------------------
      !

      IF( nn_timing == 1 )  CALL timing_start('wad_lmt')

      IF(ln_wad) THEN

        CALL wrk_alloc( jpi, jpj, zflxp, zflxn, zflxu, zflxv, zflxu1, zflxv1 )
        CALL wrk_alloc( jpi, jpj, zwdlmtu, zwdlmtv)
        CALL wrk_alloc( jpi, jpj, iwdflg )
        !
       
        IF(lwp) WRITE(numout,*)
        IF(lwp) WRITE(numout,*) 'wad_lmt : wetting/drying limiters and velocity limiting'
       
        z2dt = 2. * rdt                                 ! Euler or leap-frog time step 
        IF( neuler == 0 .AND. kt == nit000 )  z2dt = rdt
       
        zflxp(:,:)   = 0._wp
        zflxn(:,:)   = 0._wp
        zflxu(:,:)   = 0._wp
        zflxv(:,:)   = 0._wp

        zwdlmtu(:,:)  = 1._wp
        zwdlmtv(:,:)  = 1._wp
       
        ! Horizontal Flux in u and v direction
        DO jk = 1, jpkm1  
           DO jj = 1, jpjm1
              DO ji = 1, jpim1
                 zflxu(ji,jj) = zflxu(ji,jj) + fse3u(ji,jj,jk) * un(ji,jj,jk) * umask(ji,jj,jk)
                 zflxv(ji,jj) = zflxv(ji,jj) + fse3v(ji,jj,jk) * vn(ji,jj,jk) * vmask(ji,jj,jk)
              END DO  
           END DO  
        END DO
       
        zflxu(:,:) = zflxu(:,:) * e2u(:,:)
        zflxv(:,:) = zflxv(:,:) * e1v(:,:)
       
        DO jj = 2, jpjm1
           DO ji = fs_2, fs_jpim1   ! vector opt.

              IF(tmask(ji, jj, 1) < 0.5_wp) CYCLE 

              zflxp(ji,jj) = max(zflxu(ji,jj), 0._wp) - min(zflxu(ji-1,jj),   0._wp) + &
                           & max(zflxv(ji,jj), 0._wp) - min(zflxv(ji,  jj-1), 0._wp) 
              zflxn(ji,jj) = min(zflxu(ji,jj), 0._wp) - max(zflxu(ji-1,jj),   0._wp) + &
                           & min(zflxv(ji,jj), 0._wp) - max(zflxv(ji,  jj-1), 0._wp) 

              zdep2 = bathy(ji,jj) + sshb(ji,jj) - rn_wadmin 
              IF(zdep2 < 0._wp) THEN
                zdep2 = 0._wp
                sshb(ji,jj) = rn_wadmin - bathy(ji,jj)
              END IF
           ENDDO
        END DO

      
        !! start limiter iterations 
        DO jk1 = 1, nn_waditr
       
           zflag = 0     ! flag indicating if any further iteration is needed?
          
           zflxu1(:,:) = zflxu(:,:) * zwdlmtu(:,:)
           zflxv1(:,:) = zflxv(:,:) * zwdlmtv(:,:)
          
           DO jj = 2, jpjm1
              DO ji = fs_2, fs_jpim1   ! vector opt.
        
                 IF(tmask(ji, jj, 1) < 0.5_wp) CYCLE 
        
                 ztmp = e1t(ji,jj) * e2t(ji,jj)     !there must be an array ready for this

                 zzflxp = max(zflxu1(ji,jj), 0._wp) - min(zflxu1(ji-1,jj),   0._wp) + &
                        & max(zflxv1(ji,jj), 0._wp) - min(zflxv1(ji,  jj-1), 0._wp) 
                 zzflxn = min(zflxu1(ji,jj), 0._wp) - max(zflxu1(ji-1,jj),   0._wp) + &
                        & min(zflxv1(ji,jj), 0._wp) - max(zflxv1(ji,  jj-1), 0._wp) 
          
                 zdep1 = (zzflxp + zzflxn) * z2dt / ztmp
                 zdep2 = bathy(ji,jj) + sshb(ji,jj) - rn_wadmin   ! this one can be moved out of the loops
          
                 IF(zdep1 > zdep2) THEN
                   zflag = 1
                   zcoef = ( ( zdep2 - rn_wadmine ) * ztmp - zzflxn * z2dt ) / ( zflxp(ji,jj) * z2dt )
                   zcoef = max(zcoef, 0._wp)
                   iwdflg(ji, jj) = 1
                   IF(zflxu1(ji,  jj) > 0._wp) zwdlmtu(ji  ,jj) = zcoef
                   IF(zflxu1(ji-1,jj) < 0._wp) zwdlmtu(ji-1,jj) = zcoef
                   IF(zflxv1(ji,  jj) > 0._wp) zwdlmtv(ji  ,jj) = zcoef
                   IF(zflxv1(ji,jj-1) < 0._wp) zwdlmtv(ji-1,jj) = zcoef
                 END IF
              END DO ! ji loop
           END DO  ! jj loop
          
           CALL lbc_lnk( zwdlmtu, 'U', 1. )
           CALL lbc_lnk( zwdlmtv, 'V', 1. )
          
           IF(zflag == 0) EXIT
          
           IF(jk1 == nn_waditr) THEN
              
              IF(zflxu1(ji,  jj  ) >= 0._wp) zwdlmtu(ji,  jj  ) = 0._wp
              IF(zflxu1(ji-1,jj  ) <  0._wp) zwdlmtu(ji-1,jj  ) = 0._wp
              IF(zflxv1(ji,  jj  ) >= 0._wp) zwdlmtv(ji,  jj  ) = 0._wp
              IF(zflxv1(ji,  jj-1) <  0._wp) zwdlmtv(ji,  jj-1) = 0._wp

              CALL lbc_lnk( zwdlmtu, 'U', 1. )
              CALL lbc_lnk( zwdlmtv, 'V', 1. )
           END IF

        END DO  ! jk1 loop
       
        un(:,:,:) = un(:,:,:) * zwdlmtu(ji, jj) 
        vn(:,:,:) = vn(:,:,:) * zwdlmtv(ji, jj) 

        CALL lbc_lnk( un, 'U', -1. )
        CALL lbc_lnk( vn, 'V', -1. )
       
        IF(zflag == 1 .AND. lwp) WRITE(numout,*) 'Need more iterations in wad_lmt!!!'
       
        !IF( ln_rnf      )   CALL sbc_rnf_div( hdivn )          ! runoffs (update hdivn field)
        !IF( nn_cla == 1 )   CALL cla_div    ( kt )             ! Cross Land Advection (update hdivn field)
        !
        !
        CALL wrk_dealloc( jpi, jpj, zflxp, zflxn, zflxu, zflxv, zflxu1, zflxv1 )
        CALL wrk_dealloc( jpi, jpj, zwdlmtu, zwdlmtv)
        CALL wrk_dealloc( jpi, jpj, iwdflg )
      !
      END IF

      IF( nn_timing == 1 )  CALL timing_stop('wad_lmt')
   END SUBROUTINE wad_lmt
   !!======================================================================
END MODULE wadlmt
